/* Langton's Ant
 * 
 * Author: Aaron Weeden, Shodor, 2016
 * 
 * Naming convention:
 *   ALL_CAPS for constants
 *   PascalCase for globals and functions
 *   camelCase for locals
 */

/*******************************************************************************
 * IMPORT LIBRARIES
 ******************************************************************************/
#include <netcdf.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/*******************************************************************************
 * DEFINE CONSTANTS
 ******************************************************************************/
// Model parameters
#define TIME_STEP_COUNT 1000
#define ANT_COUNT 1
#define CELL_COUNT_X 21
#define CELL_COUNT_Y CELL_COUNT_X
#define RANDOM_SEED 0

// Cell states
#define BLACK 0
#define WHITE 1

// Ant directions
#define UP    0
#define LEFT  1
#define RIGHT 2
#define DOWN  3

// Define NetCDF macros
#define FILENAME "langton.nc"
#define INDEPENDENT_VAR_COUNT 3 // time, y, x
#define TIME_NAME "Time"
#define TIME_UNITS "time steps since the simulation started"
#define Y_NAME "Y"
#define Y_UNITS "ant lengths from the bottom"
#define X_NAME "X"
#define X_UNITS "ant lengths from the left"
#define STATE_NAME "State"

/*******************************************************************************
 * DECLARE GLOBAL VARIABLES
 ******************************************************************************/
int CellStates[CELL_COUNT_Y][CELL_COUNT_X];

// Declare NetCDF ID
int NcId;

// Declare NetCDF variable IDs
int TimeVarId;
int YVarId;
int XVarId;
int StateVarId;

// Declare NetCDF functions
void TryNc(int const err);
void NcInit();
  void PutDimVals(int const varId, int const count);
void NcOutput(int const time);
void NcFinalize();

/*******************************************************************************
 * DEFINE FUNCTIONS
 ******************************************************************************/
// Generate a random integer between [min..max)
int RandBetween(const int min, const int max) {
  return min + (random() % (max - min));
}

/*******************************************************************************
 * DEFINE MAIN FUNCTION
 ******************************************************************************/
int main() {
  NcInit();

  // Initialize cell states
  for (int y = 0; y < CELL_COUNT_Y; y++) {
    for (int x = 0; x < CELL_COUNT_X; x++) {
      CellStates[y][x] = WHITE;
    }
  }

  // Seed the random number generator
  srandom(RANDOM_SEED);

  // Give each ant a position and direction
  int antCellXs[ANT_COUNT];
  int antCellYs[ANT_COUNT];
  int antDirs[ANT_COUNT];
  for (int i = 0; i < ANT_COUNT; i++) {
    antCellXs[i] = RandBetween(0, CELL_COUNT_X);
    antCellYs[i] = RandBetween(0, CELL_COUNT_Y);
    antDirs[i] = LEFT;
  }

  // Start the simulation
  for (int timeStep = 0; timeStep < TIME_STEP_COUNT; timeStep++) {
    // Output
    NcOutput(timeStep);
    printf("Time %d\n", timeStep);
    for (int y = CELL_COUNT_Y-1; y >= 0; y--) {
      for (int x = 0; x < CELL_COUNT_X; x++) {
        printf("%d", CellStates[y][x]);
      }
      printf("\n");
    }
    printf("\n");

    // Turn
    for (int i = 0; i < ANT_COUNT; i++) {
      int const isBlack = (CellStates[antCellYs[i]][antCellXs[i]] == BLACK);
      if (       (antDirs[i] == UP    &&  isBlack) ||
                 (antDirs[i] == DOWN  && !isBlack)) {
        antDirs[i] = LEFT;
      } else if ((antDirs[i] == UP    && !isBlack) ||
                 (antDirs[i] == DOWN  &&  isBlack)) {
        antDirs[i] = RIGHT;
      } else if ((antDirs[i] == LEFT  &&  isBlack) ||
                 (antDirs[i] == RIGHT && !isBlack)) {
        antDirs[i] = DOWN;
      } else if ((antDirs[i] == LEFT  && !isBlack) ||
                 (antDirs[i] == RIGHT &&  isBlack)) {
        antDirs[i] = UP;
      }
    }

    // Update cells
    for (int i = 0; i < ANT_COUNT; i++) {
      CellStates[antCellYs[i]][antCellXs[i]] =
        (CellStates[antCellYs[i]][antCellXs[i]] == BLACK) ? WHITE : BLACK;
    }

    // Move
    for (int i = 0; i < ANT_COUNT; i++) {
      switch (antDirs[i]) {
        case UP:
          if (antCellYs[i] == 0) {
            antCellYs[i] = CELL_COUNT_Y-1;
          } else {
            antCellYs[i]--;
          }
          break;
        case LEFT:
          if (antCellXs[i] == 0) {
            antCellXs[i] = CELL_COUNT_X-1;
          } else {
            antCellXs[i]--;
          }
          break;
        case RIGHT:
          if (antCellXs[i] == CELL_COUNT_X-1) {
            antCellXs[i] = 0;
          } else {
            antCellXs[i]++;
          }
          break;
        case DOWN:
          if (antCellYs[i] == CELL_COUNT_Y-1) {
            antCellYs[i] = 0;
          } else {
            antCellYs[i]++;
          }
          break;
      }
    }
  }

  NcFinalize();
  return 0;
}

// Define a function that checks the return of a NetCDF function and, if there
// is an error, prints the error and exits
void TryNc(int const err) {
  if (err) {
    fprintf(stderr, "%s\n", nc_strerror(err));
    exit(EXIT_FAILURE);
  }
}

// Define a function for initializing NetCDF
void NcInit() {
  // Create the NetCDF dataset
  TryNc(nc_create(FILENAME, NC_CLOBBER, &NcId));

  // Declare dimension IDs for each independent variable
  int dimIds[INDEPENDENT_VAR_COUNT];

  // Define NetCDF dimensions for each independent variable
  TryNc(nc_def_dim(NcId, TIME_NAME, TIME_STEP_COUNT, &(dimIds[0])));
  TryNc(nc_def_dim(NcId, Y_NAME,    CELL_COUNT_Y,    &(dimIds[1])));
  TryNc(nc_def_dim(NcId, X_NAME,    CELL_COUNT_X,    &(dimIds[2])));

  // Define NetCDF variables for each variable
  TryNc(nc_def_var(NcId, TIME_NAME,  NC_INT, 1, &(dimIds[0]), &TimeVarId));
  TryNc(nc_def_var(NcId, Y_NAME,     NC_INT, 1, &(dimIds[1]), &YVarId));
  TryNc(nc_def_var(NcId, X_NAME,     NC_INT, 1, &(dimIds[2]), &XVarId));
  TryNc(nc_def_var(NcId, STATE_NAME, NC_INT, INDEPENDENT_VAR_COUNT, dimIds,
                    &StateVarId));

  // Define NetCDF units text for each variable
  TryNc(nc_put_att_text(NcId, TimeVarId,  "units", strlen(TIME_UNITS),
                         TIME_UNITS));
  TryNc(nc_put_att_text(NcId, YVarId,     "units", strlen(Y_UNITS),
                         Y_UNITS));
  TryNc(nc_put_att_text(NcId, XVarId,     "units", strlen(X_UNITS),
                         X_UNITS));

  // Leave NetCDF define mode
  TryNc(nc_enddef(NcId));

  // Write the values for the each dimension
  PutDimVals(TimeVarId, TIME_STEP_COUNT);
  PutDimVals(YVarId, CELL_COUNT_Y);
  PutDimVals(XVarId, CELL_COUNT_X);
}

// Define a function that outputs values for a NetCDF dimension's variable
void PutDimVals(int const varId, int const count) {
  // Fill a 1D array with the numbers 0 through count
  int vals[count];
  for (int i = 0; i < count; i++) {
    vals[i] = i;
  }

  // Write the array to the given NetCDF variable
  TryNc(nc_put_var_int(NcId, varId, vals));
}

// Define function to output NetCDF dependent variables
void NcOutput(int const time) {
  // Declare where to start writing data for each dimension
  size_t const startArray[INDEPENDENT_VAR_COUNT] = {
    time, // Start at the current time step
    0, 0 // Start at coordinate (0, 0)
  };

  // Declare how much data to write in each dimension
  size_t const countArray[INDEPENDENT_VAR_COUNT] = {
    1, // Write one time step of data
    CELL_COUNT_Y, CELL_COUNT_X // Write all positions in the
                               // environment
  };

  // Output the C array into the NetCDF variable
  TryNc(nc_put_vara_int(NcId, // NetCDF dataset ID
                         StateVarId, // NetCDF variable ID
                         startArray, // Where to start writing data in each
                                     // dimension
                         countArray, // How much data to write in each
                                     // dimension
                         &(CellStates[0][0]))); // C array of data to output to
                                                // the NetCDF variable
}

// Close the NetCDF dataset
void NcFinalize() {
  TryNc(nc_close(NcId));
}

